//
//	bmp.cpp - source file / freeware
//
//	David Henry - tfc_duke@club-internet.fr
//


#ifdef _WIN32
#define	WIN32_LEAN_AND_MEAN
#include	<windows.h>
#endif

#include	<GL/gl.h>
#include	<fstream>

#include	"texture.h"


// identifiant "BM"
#define	BITMAP_ID			('B' + ('M'<<8))

// status premier/second octet pour compression RLE
#define	RLE_COMMAND			0
#define	RLE_ENDOFLINE		0
#define	RLE_ENDOFBITMAP		1
#define	RLE_DELTA			2

#define	BMP_WIN				0
#define	BMP_OS2				1


#ifndef _WIN32

// type de compression
#define	BI_RGB				0
#define	BI_RLE8				1
#define	BI_RLE4				2
#define	BI_BITFIELDS		3


// --------------------------------------------
// BITMAPFILEHEADER - bitmap file header.
// --------------------------------------------

#pragma pack(2)
typedef struct tagBITMAPFILEHEADER		// bmfh
{
	unsigned short	bfType;				// numro magique "BM"
	unsigned int	bfSize;				// taille du fichier
	unsigned short	bfReserved1;		// rserv
	unsigned short	bfReserved2;		// rserv
	unsigned int	bfOffBits;			// offset pour les donnes bitmap

} BITMAPFILEHEADER, *PBITMAPFILEHEADER;
#pragma pack(4)



// --------------------------------------------
// BITMAPINFOHEADER - bitmap info header.
// --------------------------------------------

typedef struct
{
	unsigned int	biSize;				// taille de la structure
	int				biWidth;			// largeur de l'image
	int				biHeight;			// hauteur de l'mage
	unsigned short	biPlanes;			// doit tre gal  1
	unsigned short	biBitCount;			// nombre de bits par pixels
	unsigned int	biCompression;		// type de compression
	unsigned int	biSizeImage;		// taille des donnes pixels
	int				biXPelsPerMeter;	// nombre de pixels par mtre sur l'axe des x
	int				biYPelsPerMeter;	// nombre de pixels par mtre sur l'axe des y
	unsigned int	biClrUsed;			// nombre de couleur utilises par le bitmap
	unsigned int	biClrImportant;		// nombre de couleurs importantes

} BITMAPINFOHEADER;



// --------------------------------------------
// BITMAPCOREHEADER - bitmap core header.
// --------------------------------------------

typedef struct tagBITMAPCOREHEADER		// bmch
{
	unsigned int	bcSize;				// taille de la structure
	unsigned short	bcWidth;			// doit tre gal  1
	unsigned short	bcHeight;			// nombre de bits par pixels
	unsigned short	bcPlanes;			// doit tre gal  1
	unsigned short	bcBitCount;			// nombre de bits par pixels

} BITMAPCOREHEADER, *PBITMAPCOREHEADER;



// --------------------------------------------
// RGBTRIPLE - pixel 24 bits.
// --------------------------------------------

typedef struct tagRGBTRIPLE			// rgbt
{
	unsigned char	rgbtBlue;
	unsigned char	rgbtGreen;
	unsigned char	rgbtRed;

} RGBTRIPLE, *PRGBTRIPLE;



// --------------------------------------------
// RGBQUAD - pixel 32 bits.
// --------------------------------------------

typedef struct tagRGBQUAD			// rgbt
{
	unsigned char	rgbBlue;
	unsigned char	rgbGreen;
	unsigned char	rgbRed;
	unsigned char	rgbReserved;

} RGBQUAD, *PRGBQUAD;



// --------------------------------------------
// BITMAPCOREINFO - bitmap core info.
// --------------------------------------------

typedef struct tagBITMAPCOREINFO		// bmci
{
	BITMAPCOREHEADER	bmciHeader;		// taille de la structure
	RGBTRIPLE			bcmiColors[1];	// doit tre gal  1

} BITMAPCOREINFO, *PBITMAPCOREINFO;

#endif	// WIN32



// --------------------------------------------------
// LoadFileBMP() - charge une image BMP. Si bFlipVert
// vaut 'true', l'image est retourne verticalement.
// formats accepts:
//     # BGR 24-32 bits
//     # index couleur 1-4-8 bits
//     # index couleur 4-8 bits (compression RLE)
// --------------------------------------------------

bool LoadFileBMP( const std::string &szFileName, gl_texture_t **img, bool bFlipVert )
{
	std::ifstream		file;					// fichier
	BITMAPFILEHEADER	*bmfh = NULL;			// bitmap file header
	BITMAPINFOHEADER	*bmih = NULL;			// bitmap info header (windows)
	BITMAPCOREHEADER	*bmch = NULL;			// bitmap core header (os/2)
	RGBTRIPLE			*os2_palette = NULL;	// pointeur palette couleurs os/2
	RGBQUAD				*win_palette = NULL;	// pointeur palette couleurs windows
	char				*buffer = NULL;			// tampon mmoire stockant les donnes du fichier
	unsigned char		*ptr = NULL;			// pointeur donnes pixels
	unsigned int		bitCount;				// nombre de bit par pixel
	unsigned int		compression;			// type de compression (rgb/rle)
	int					os_type;				// bitmap Windows ou OS/2
	int					row, col, i;			// variables de contrle



	/////////////////////////////////////////////////////
	// lecture de l'intgralit du  fichier

	file.open( szFileName.c_str(), std::ios::in | std::ios::binary );

	if( file.fail() )
		return false;

	file.seekg( 0, std::ios::end );
	long flen = file.tellg();
	file.seekg( 0, std::ios::beg );

	buffer = new char[ flen ];
	file.read( buffer, flen );
	char *pBuff = buffer;

	file.close();

	/////////////////////////////////////////////////////


	// lecture du header
	bmfh = (BITMAPFILEHEADER *)pBuff;
	pBuff += sizeof( BITMAPFILEHEADER );

	// on vrifie qu'il s'agit bien d'un BITMAP
	if( bmfh->bfType != BITMAP_ID )
	{
		delete [] buffer;
		return false;
	}

	// infos en-tte du bitmap
	bmch = (BITMAPCOREHEADER *)pBuff;
	bmih = (BITMAPINFOHEADER *)pBuff;

	// cre un objet texture
	*img = new gl_texture_t;
	(*img)->format = GL_RGB;
	(*img)->internalFormat = 3;

	if( bmih->biCompression > 3 )
	{
		// OS/2 style
		pBuff += sizeof( BITMAPCOREHEADER );

		bitCount = bmch->bcBitCount;
		compression = BI_RGB;
		os_type = BMP_OS2;

		(*img)->width = bmch->bcWidth;
		(*img)->height = bmch->bcHeight;
	}
	else
	{
		// WINDOWS style
		pBuff += sizeof( BITMAPINFOHEADER );

		bitCount = bmih->biBitCount;
		compression = bmih->biCompression;
		os_type = BMP_WIN;

		(*img)->width = bmih->biWidth;
		(*img)->height = bmih->biHeight;
	}


	/////////////////////////////////////////////////////
	// lecture de la palette

	if( bitCount <= 8 )
	{
		// les images 16, 24 et 32 bits ne possdent pas de palettes.

		// on ajuste le pointeur de la palette sur la zone mmoire
		// correspondante dans buffer
		os2_palette = (RGBTRIPLE *)pBuff;
		win_palette = (RGBQUAD *)pBuff;

		// [nombre de couleurs palette] * [taille d'un pixel]
		pBuff += (1 << bitCount) * (bitCount >> 3) * sizeof( unsigned char );
	}
	else
	{
		os2_palette = NULL;
		win_palette = NULL;
	}

	/////////////////////////////////////////////////////


	// allocation de mmoire pour les donnes pixels de l'image
	(*img)->texels = new GLubyte[ (*img)->width * (*img)->height * (*img)->internalFormat ];
	ptr = &(*img)->texels[0];

	// on dplace le pointeur fichier au dbut des donnes pixels
	pBuff = buffer + (bmfh->bfOffBits * sizeof( unsigned char ));


	/////////////////////////////////////////////////////
	// lecture des donnes suivant le type de
	// compression et le nombre de bits par pixel
	/////////////////////////////////////////////////////

	switch( compression )
	{
		case BI_RGB:
		{
			for( row = (*img)->height - 1; row >= 0; row-- )
			{
				if( bFlipVert )
					ptr = &(*img)->texels[ row * (*img)->width * (*img)->internalFormat ];

				switch( bitCount )
				{
					case 1:
					{
// RGB 1 BITS
						for( col = 0; col < (int)((*img)->width / 8); col++ )
						{
							// lecture du pixel en cours
							unsigned char color = *((unsigned char *)(pBuff++));

							for( i = 7; i >= 0; i--, ptr += 3 )
							{
								// conversion pixel index couleur bgr 1 bit
								// en pixel rgb 24 bits
								int clrIdx = ((color & (1<<i)) > 0);

								if( os_type == BMP_OS2 )
								{
									ptr[0] = (GLubyte)os2_palette[ clrIdx ].rgbtRed;
									ptr[1] = (GLubyte)os2_palette[ clrIdx ].rgbtGreen;
									ptr[2] = (GLubyte)os2_palette[ clrIdx ].rgbtBlue;
								}
								else
								{
									ptr[0] = (GLubyte)win_palette[ clrIdx ].rgbRed;
									ptr[1] = (GLubyte)win_palette[ clrIdx ].rgbGreen;
									ptr[2] = (GLubyte)win_palette[ clrIdx ].rgbBlue;
								}
							}
						}

						break;
					}

					case 4:
					{
// RGB 4 BITS
						for( col = 0; col < (int)((*img)->width / 2); col++, ptr += 6 )
						{
							// lecture du pixel en cours
							unsigned char color = *((unsigned char *)(pBuff++));

							// conversion pixel index couleur bgr 4 bits
							// en pixel rgb 24 bits
							int clrIdx;

							if( os_type == BMP_OS2 )
							{
								clrIdx = (color >> 4);
								ptr[0] = (GLubyte)os2_palette[ clrIdx ].rgbtRed;
								ptr[1] = (GLubyte)os2_palette[ clrIdx ].rgbtGreen;
								ptr[2] = (GLubyte)os2_palette[ clrIdx ].rgbtBlue;

								clrIdx = (color & 0x0F);
								ptr[3] = (GLubyte)os2_palette[ clrIdx ].rgbtRed;
								ptr[4] = (GLubyte)os2_palette[ clrIdx ].rgbtGreen;
								ptr[5] = (GLubyte)os2_palette[ clrIdx ].rgbtBlue;
							}
							else
							{
								clrIdx = (color >> 4);
								ptr[0] = (GLubyte)win_palette[ clrIdx ].rgbRed;
								ptr[1] = (GLubyte)win_palette[ clrIdx ].rgbGreen;
								ptr[2] = (GLubyte)win_palette[ clrIdx ].rgbBlue;

								clrIdx = (color & 0x0F);
								ptr[3] = (GLubyte)win_palette[ clrIdx ].rgbRed;
								ptr[4] = (GLubyte)win_palette[ clrIdx ].rgbGreen;
								ptr[5] = (GLubyte)win_palette[ clrIdx ].rgbBlue;
							}
						}

						break;
					}

					case 8:
					{
// RGB 8 BITS
						for( col = 0; col < (int)(*img)->width; col++, ptr += 3 )
						{
							// lecture du pixel en cours
							unsigned char color = *((unsigned char *)(pBuff++));

							// conversion pixel index couleur bgr 8 bits
							// en pixel rgb 24 bits
							if( os_type == BMP_OS2 )
							{
								ptr[0] = (GLubyte)os2_palette[ color ].rgbtRed;
								ptr[1] = (GLubyte)os2_palette[ color ].rgbtGreen;
								ptr[2] = (GLubyte)os2_palette[ color ].rgbtBlue;
							}
							else
							{
								ptr[0] = (GLubyte)win_palette[ color ].rgbRed;
								ptr[1] = (GLubyte)win_palette[ color ].rgbGreen;
								ptr[2] = (GLubyte)win_palette[ color ].rgbBlue;
							}
						}

						break;
					}

					case 24:
					{
// RGB 24 BITS
						for( col = 0; col < (int)((*img)->width); col++, ptr += 3 )
						{
							// lecture du pixel 24 bits et conversion bgr->rgb
							RGBTRIPLE *pix = (RGBTRIPLE *)pBuff;
							pBuff += sizeof( RGBTRIPLE );

							ptr[0] = (GLubyte)pix->rgbtRed;
							ptr[1] = (GLubyte)pix->rgbtGreen;
							ptr[2] = (GLubyte)pix->rgbtBlue;
						}

  						break;
					}

					case 32:
					{
// RGB 32 BITS
						// le quatrime octet ne sert pas, il s'agit donc de
						// pixels 24 bits stocks sur 32 bits
						for( col = 0; col < (int)((*img)->width); col++, ptr += 3 )
						{
							// lecture du pixel 32 bits et conversion bgr->rgb
							RGBQUAD *pix = (RGBQUAD *)(pBuff += sizeof( RGBQUAD ));
							pBuff += sizeof( RGBQUAD );

							ptr[0] = (GLubyte)pix->rgbRed;
							ptr[1] = (GLubyte)pix->rgbGreen;
							ptr[2] = (GLubyte)pix->rgbBlue;
						}

						break;
					}
				}
			}

			break;
		}

		case BI_RLE8:
		{
// RLE 8 BITS
			for( row = (*img)->height - 1; row >= 0; row-- )
			{
				if( bFlipVert )
					ptr = &(*img)->texels[ row * (*img)->width * (*img)->internalFormat ];

				for( col = 0; col < (int)((*img)->width); /* rien */ )
				{
					// on rcupre une unit de code (sur 2 octets)
					unsigned char byte1 = *((unsigned char *)(pBuff++));
					unsigned char byte2 = *((unsigned char *)(pBuff++));

					if( byte1 == RLE_COMMAND )
					{
						// encodage absolu
						for( i = 0; i < byte2; i++, ptr += 3, col++ )
						{
							// lecture du pixel en cours
							unsigned char color = *((unsigned char *)(pBuff++));

							// conversion pixel index couleur bgr 8 bits
							// en pixel rgb 24 bits
							ptr[0] = (GLubyte)win_palette[ color ].rgbRed;
							ptr[1] = (GLubyte)win_palette[ color ].rgbGreen;
							ptr[2] = (GLubyte)win_palette[ color ].rgbBlue;
						}

						// on avance d'un octet si l'on s'est arrt
						// sur un nombre impair d'octets dcods
						if( (byte2 % 2) == 1 )
							pBuff++;
					}
					else
					{
						// on decode pour les 'byte1' pixels suivants
						for( i = 0; i < byte1; i++, ptr += 3, col++ )
						{
							// conversion pixel index couleur bgr 8 bits
							// en pixel rgb 24 bits
							ptr[0] = (GLubyte)win_palette[ byte2 ].rgbRed;
							ptr[1] = (GLubyte)win_palette[ byte2 ].rgbGreen;
							ptr[2] = (GLubyte)win_palette[ byte2 ].rgbBlue;
						}        
					}
				}
			}

			break;
		}

		case BI_RLE4:
		{
// RLE 4 BITS
			unsigned char color;
			int	bytesRead = 0;		// nombre d'octets lus du fichier

			for( row = (*img)->height - 1; row >= 0; row-- )
			{
				if( bFlipVert )
					ptr = &(*img)->texels[ row * (*img)->width * (*img)->internalFormat ];

				for( col = 0; col < (int)((*img)->width); /* rien */ )
				{
					// on rcupre une unit de code (sur 2 octets)
					unsigned char byte1 = *((unsigned char *)(pBuff++));
					unsigned char byte2 = *((unsigned char *)(pBuff++));

					bytesRead += 2;
      
					if( byte1 == RLE_COMMAND )
					{
						// encodage absolu
						unsigned char databyte = 0;

						for( i = 0; i < byte2; i++, ptr += 3, col++ )
						{
							if( (i % 2) == 0 )
							{
								// lecture du pixel en cours
								databyte = *((unsigned char *)(pBuff++));
								bytesRead++;

								color = (databyte >> 4);	// 4 premiers bits
							}
							else
							{
								color = (databyte & 0x0F);	// 4 derniers bits
							}

							// conversion pixel index couleur bgr 4 bits
							// en pixel rgb 24 bits
							ptr[0] = (GLubyte)win_palette[ color ].rgbRed;
							ptr[1] = (GLubyte)win_palette[ color ].rgbGreen;
							ptr[2] = (GLubyte)win_palette[ color ].rgbBlue;
						}
          
						// on avance d'un octet si l'on s'est arrt
						// sur un nombre impair d'octets dcods
						while( (bytesRead % 2) != 0 )
						{
							pBuff++;
							bytesRead++;
						}
					}
					else
					{
						// on decode pour les 'byte1' pixels suivants
						for( i = 0; i < byte1; i++, ptr += 3, col++ )
						{
							if( (i % 2) == 0 )
								color = (byte2 >> 4);	// 4 premiers bits
							else
								color = (byte2 & 0x0F);	// 4 derniers bits

							// conversion pixel index couleur bgr 4 bits
							// en pixel rgb 24 bits
							ptr[0] = (GLubyte)win_palette[ color ].rgbRed;
							ptr[1] = (GLubyte)win_palette[ color ].rgbGreen;
							ptr[2] = (GLubyte)win_palette[ color ].rgbBlue;
						}        
					}
				}
			}

			break;
		}

		default:
		{
			// format inconnu ou non trat
			delete [] buffer;
			delete *img;
			return true;
		}
	}

	delete [] buffer;
	return true;
}
