//
//	main.cpp - source file
//
//	David Henry - tfc_duke@club-internet.fr
//


#ifdef _WIN32
#define	WIN32_LEAN_AND_MEAN
#include	<windows.h>
// Microsoft Visual C++
#ifdef _MSC_VER
#pragma comment( lib, "opengl32.lib" )
#pragma comment( lib, "glu32.lib" )
#pragma comment( lib, "SDL.lib" )
#pragma comment( lib, "SDLmain.lib" )
#endif	// _MSC_VER
#endif	// _WIN32

#include	<GL/gl.h>
#include	<GL/glu.h>
#include	<iostream>
#include	<iomanip>
#include	<stdio.h>
#include	<SDL.h>

#include	"texture.h"

using namespace std;




/////////////////////////////////////////////////
// texture ID
GLuint	tex = 1;
CTextureManager *pTexMgr;

// surface SDL
SDL_Surface *surface;
const char title[] = "OpenGL Texture Manager Demo";

// camra
struct vector3d
{
	float	x, y, z;

} rot, eye;
/////////////////////////////////////////////////



// ----------------------------------------------
// Shutdown() - fin du programme.
// ----------------------------------------------

void Shutdown( int returnCode )
{
	CTextureManager::Kill();

	// clean up the window
	SDL_Quit();

	// quitte proprement
	exit( returnCode );
}



// ----------------------------------------------
// Initialize() - initialisation d'opengl.
// ----------------------------------------------

void Initialize( void )
{
	glClearColor( 0.5f, 0.5f, 0.5f, 0.0f );
	glShadeModel( GL_SMOOTH );
	glEnable( GL_DEPTH_TEST );

	/////////////////////////////////////////////

	// initialisation camra
	rot.x = 0.0;		eye.x = 0.0;
	rot.y = 0.0;		eye.y = 0.0;
	rot.z = 0.0;		eye.z = 3.0;

	// chargement des textures
	pTexMgr = CTextureManager::GetInstance();
	pTexMgr->Initialize();

	pTexMgr->LoadTexture( "data/BMP_1bit.bmp" );
	pTexMgr->LoadTexture( "data/BMP_4bits.bmp" );
	pTexMgr->LoadTexture( "data/BMP_4bits_rle.bmp" );
	pTexMgr->LoadTexture( "data/BMP_8bits.bmp" );
	pTexMgr->LoadTexture( "data/BMP_8bits_rle.bmp" );
	pTexMgr->LoadTexture( "data/BMP_24bits.bmp" );
	pTexMgr->LoadTexture( "data/BMP_1bit_os2.bmp" );
	pTexMgr->LoadTexture( "data/BMP_4bits_os2.bmp" );
	pTexMgr->LoadTexture( "data/BMP_8bits_os2.bmp" );
	pTexMgr->LoadTexture( "data/BMP_24bits_os2.bmp" );

	pTexMgr->LoadTexture( "data/JPEG_24bits_quality100.jpg" );
	pTexMgr->LoadTexture( "data/JPEG_24bits_quality50.jpg" );
	pTexMgr->LoadTexture( "data/JPEG_24bits_quality25.jpg" );
	pTexMgr->LoadTexture( "data/JPEG_grayscale_quality85.jpg" );

	pTexMgr->LoadTexture( "data/PCX_1bit.pcx" );
	pTexMgr->LoadTexture( "data/PCX_4bits.pcx" );
	pTexMgr->LoadTexture( "data/PCX_8bits.pcx" );
	pTexMgr->LoadTexture( "data/PCX_24bits.pcx" );

	pTexMgr->LoadTexture( "data/PNG_grayscale.png" );
	pTexMgr->LoadTexture( "data/PNG_grayscale_alpha.png" );
	pTexMgr->LoadTexture( "data/PNG_8bits.png" );
	pTexMgr->LoadTexture( "data/PNG_8bits_alpha.png" );
	pTexMgr->LoadTexture( "data/PNG_24bits.png" );
	pTexMgr->LoadTexture( "data/PNG_32bits.png" );

	pTexMgr->LoadTexture( "data/TGA_8bits.tga" );
	pTexMgr->LoadTexture( "data/TGA_8bits_rle.tga" );
	pTexMgr->LoadTexture( "data/TGA_16bits.tga" );
	pTexMgr->LoadTexture( "data/TGA_16bits_rle.tga" );
	pTexMgr->LoadTexture( "data/TGA_24bits.tga" );
	pTexMgr->LoadTexture( "data/TGA_24bits_rle.tga" );
	pTexMgr->LoadTexture( "data/TGA_32bits.tga" );
	pTexMgr->LoadTexture( "data/TGA_32bits_rle.tga" );
	pTexMgr->LoadTexture( "data/TGA_grayscale_rle.tga" );

	/////////////////////////////////////////////
}



// ----------------------------------------------
// Resize() - rinitialisation de la fentre
// lors d'un redimensionnement...
// ----------------------------------------------

void Resize( GLint width, GLint height )
{
	// on vite les divisions par zro...
	if( height == 0 )
		height = 1;

	glViewport( 0, 0, width, height );

	// on rinitialise la matrice de projection
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( 45.0, (GLfloat)width/(GLfloat)height, 0.1, 1000.0 );

	// on rinitialise la matrice de modlisation/visualisation
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
}



// ----------------------------------------------
// Display() - redessine la scne opengl.
// ----------------------------------------------

void Display( void )
{
	// on nettoie l'cran pour redesiner la scne
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	glLoadIdentity();


	// rotation camra
	glTranslatef( -eye.x, -eye.y, -eye.z );
	glRotatef( rot.x, 1.0f, 0.0f, 0.0f );
	glRotatef( rot.y, 0.0f, 1.0f, 0.0f );
	glRotatef( rot.z, 0.0f, 0.0f, 1.0f );


	glEnable( GL_TEXTURE_2D );
	glBindTexture( GL_TEXTURE_2D, tex );

	glEnable( GL_BLEND );
	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

	// dessin d'un carr
	glBegin( GL_QUADS );
		glTexCoord2f( 0.0f, 0.0f );
		glVertex3f( -1.0f, -1.0f, 0.0f );

		glTexCoord2f( 1.0f, 0.0f );
		glVertex3f( 1.0f, -1.0f, 0.0f );

		glTexCoord2f( 1.0f, 1.0f );
		glVertex3f( 1.0f, 1.0f, 0.0f );

		glTexCoord2f( 0.0f, 1.0f );
		glVertex3f( -1.0f, 1.0f, 0.0f );
	glEnd();

	glDisable( GL_TEXTURE_2D );
}



// ----------------------------------------------
// Keyboard() - capture clavier.
// ----------------------------------------------

void handleKeyPress( SDL_keysym *keysym )
{
	switch( keysym->sym )
	{
		case SDLK_KP_MINUS:
			if( (--tex) <= 0 )
				tex = 1;

			break;

		case SDLK_KP_PLUS:
			if( (++tex) > pTexMgr->GetTextureCount() )
				tex = pTexMgr->GetTextureCount();

			break;

		case SDLK_F1:
			SDL_WM_ToggleFullScreen( surface );
			break;

		case SDLK_ESCAPE:
			Shutdown(0);
			break;

		default:
			break;
	}
}



// ----------------------------------------------
// MouseMove() - contrle souris.
// ----------------------------------------------

void MouseMove( int x, int y )
{
	static int x_pos = 0;
	static int y_pos = 0;

	SDLMod km = SDL_GetModState();

	if( SDL_GetMouseState( NULL, NULL ) & SDL_BUTTON_RMASK ) // right
	{
		// zoom
		eye.z += (x - x_pos) * 0.1;
	}
	else if( SDL_GetMouseState( NULL, NULL ) & SDL_BUTTON_LMASK ) // left
	{
		if( (km & KMOD_RSHIFT) || (km & KMOD_LSHIFT))
		{
			// translation
			eye.x -= (x - x_pos) * 0.1;
			eye.y += (y - y_pos) * 0.1;
		}
		else
		{
			// rotation
			rot.x += (y - y_pos);
			rot.y += (x - x_pos);
		}
	}

	x_pos = x;
	y_pos = y;
}



// ----------------------------------------------
// main() - fonction principale.
// ----------------------------------------------

int main( int argc, char *argv[] )
{
	cout << "OpenGL Texture Manager Demo // by David Henry : tfc_duke@club-internet.fr" << endl;
	cout << "Built release " << __DATE__ << " at " << __TIME__ << endl;

	int videoFlags;
	bool done = false;
	SDL_Event event;
	const SDL_VideoInfo *videoInfo;
	bool isActive = true;


	// initialisation de SDL
	if( SDL_Init( SDL_INIT_VIDEO ) < 0 )
	{
		cout << "Video initialization failed: " << SDL_GetError() << endl;
		Shutdown(1);
	}

	// on rcupre les infos video
	videoInfo = SDL_GetVideoInfo();

	if( !videoInfo )
	{
		cout << "Video query failed: " << SDL_GetError() << endl;
		Shutdown(1);
	}

	// flags  passer  SDL_SetVideoMode
	videoFlags  = SDL_OPENGL;			// utilise OpenGL avec SDL
	videoFlags |= SDL_GL_DOUBLEBUFFER;	// utilise le double buffering
	videoFlags |= SDL_HWPALETTE;		// pas d'approximations pour les couleurs en mode 8 bits
	videoFlags |= SDL_RESIZABLE;		// accepte le redimensionnement de la fentre

	// check si la surface peut-tre alloue en mmoire video
	if( videoInfo->hw_available )
		videoFlags |= SDL_HWSURFACE;
	else
		videoFlags |= SDL_SWSURFACE;

    // check si hardware blits possible
	if( videoInfo->blit_hw )
		videoFlags |= SDL_HWACCEL;

	// initialisation du double buffering pour OpenGL
	SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );

	// initialise un mode video
	surface = SDL_SetVideoMode( 250, 250, 24, videoFlags );
	SDL_WM_SetCaption( title, title );


	// vrifie que la surface n'est pas nulle
	if( !surface )
	{
		cout << "Video mode set failed: " << SDL_GetError() << endl;
		Shutdown(1);
	}

	// active la rptition des touches clavier
	if( (SDL_EnableKeyRepeat( 100, 100 /*SDL_DEFAULT_REPEAT_INTERVAL*/ ) ) )
	{
		cout << "Setting keyboard repeat failed: " << SDL_GetError() << endl;
		Shutdown(1);
	}

	// initialise OpenGL
	Initialize();

	// redimensionne la fentre initiale
	Resize( 250, 250 );

	// boucle principale (jusqu' la fin du programme)
	while( !done )
	{
		// traite les vnements SDL
		while( SDL_PollEvent( &event ) )
		{
			switch( event.type )
			{
				case SDL_ACTIVEEVENT:
				{
					// le focus de la fentre a t altr (fentre iconifie ou maximise)
					// on ne dessine plus la scne OpenGL
					if( (event.active.gain == 0) && (event.active.state & SDL_APPACTIVE) )
						isActive = false;
					else
						isActive = true;

					break;
				}

				case SDL_VIDEORESIZE:
				{
					// redimensionnement de la fentre
					surface = SDL_SetVideoMode( event.resize.w, event.resize.h, 24, videoFlags );

					if( !surface )
					{
						cout << "Could not get a surface after resize: " << SDL_GetError() << endl;
						Shutdown( 1 );
					}

					Resize( event.resize.w, event.resize.h );
					break;
				}

				case SDL_KEYDOWN:
				{
					// une touche a t presse
					handleKeyPress( &event.key.keysym );
					break;
				}

				case SDL_MOUSEMOTION:
				{
					// mouvement de la souris
					MouseMove( event.button.x, event.button.y );
					break;
				}

				case SDL_QUIT:
				{
					// fin du programme
					done = true;
					break;
				}

				default:
					break;
			}
		}

		// si la fentre est active
		if( isActive )
		{
			// dessine la scne
			Display();

			// change les tampons avant et arrire
			SDL_GL_SwapBuffers();
		}
	}

	// dtruit les objets et quitte le programme
	Shutdown(0);

	// ne devrait jamais tre excut (cf: Shutdown() appelle exit())
	return 0;
}
